/*
Module: 	xMoveIt.c
Author:		Ralph Betza
Date:		JFK anniversary, 1991
Purpose:	Is your clock burning out the phosphors in one spot of
			the screen? Use xMoveIt to move it around!

The framework for this program was swiped from xcursor.c

I use this program on tclock.
	I recommend the following changes to tclock:
	1.	Fix it so it sleeps 60 seconds if the user has specified "no
		second hand"
			-xrm 'Tclock.Second.Tip.Radius: 0' \
	2.	Fix it so it raises itself every 60 seconds.
		The beauty of tclock is that you can still read your xterm
		window even if tclock is on top of it.

Then, run tclock and xMoveIt as follows:
	tclock -ticksize 5 -borderWidth 1 -scale .5 \
		-border LimeGreen -bg tomato \
		-xrm 'Tclock.Second.Tip.Radius: 0' \
		&
	sleep 5
	xMoveIt -name xclock &

Of course, you can use any window as the target....

You could also use one window as the hour hand and another as the
minute hand, by modifying this program to circle the screen once an
hour, and running two different xMoveIt programs, one for hour hand,
one for minute hand.
	See helpful comment at the seting of Distance.

*/

#include <stdio.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/cursorfont.h>

Display *display;
Window window;
int screen;

void local_help();
Window window_by_name(), point_to_window();

#include <time.h>
#include <sys/time.h>

/* Text for select by pointer case */
char *point_txt[] = {
"           ====> Please select the window",
"           ====> you wish to have moved",
"           ====> change by clicking the",
"           ====> pointer in that window.",
0};


main(argc,argv)
int argc;
char **argv;
{
    register int opt;		/* Parsed option letter */
    char *id_str = 0;		/* Window id in ASCII */
    char *disp_str = 0;		/* Display name */
    char *name_str = 0;		/* Window name */
    register int uflag = 0;	/* Source specified */
    register char **ptr;	/* Fast pointer */
    extern int optind;
    extern char *optarg;
    extern int opterr;

    opterr = 0;			/* No error msgs from getopt */

    while ((opt = getopt(argc,argv,"d:n:i")) != -1)
      {
	  switch(opt)
	    {
	      case 'd':		/* Get display name */
		disp_str = argv[optind++];
		if (*disp_str == '-') /* Check to see if user skipped */
				      /* required data */
		  local_help();
		break;

	      case 'n':		/* Get window name */
		name_str = argv[optind++];
		if (*name_str == '-')
		  local_help();
		uflag++;
		break;

	      case 'i':		/* Get window id */
		id_str = argv[optind++];
		if (*id_str == '-')
		  local_help();
		uflag++;
		break;

	      case '?':		/* Ooops! */
		fprintf(stderr,"xMoveIt - bad option\n\n");
		local_help();
		break;
	    }
      }

    if ((display = XOpenDisplay(disp_str)) == NULL)
      {
	  fprintf(stderr,"Cannot open display %s\n",disp_str);
	  exit(1);
      }

    screen = DefaultScreen(display);

    if (name_str != NULL)
      {
	  if ((window = window_by_name(DefaultRootWindow(display),name_str))
	       == 0)
	      {
		  fprintf(stderr,"No window named %s found.\n",name_str);
		  exit(2);
	      }
      }
    else if (id_str != NULL)
      window = (Window)strtol(id_str,(char **)NULL,0);
    else if (uflag)
      window = DefaultRootWindow(display);
    else
	{	for (ptr = point_txt; *ptr; ptr++)
			printf("%s\n",*ptr);
		window = point_to_window(); /* use pointer to get window */
	}

	{	/* The code for xMoveIt starts here, and continues to the
		** matching squiggly bracket. Everything else comes from
		** xcursor.
		*/

		int w_height, w_width;
		int d_height, d_width;
		XWindowAttributes watt;
		struct minutes_ { int x; int y; } places[ 12 ][ 60 ];
		int hour, minute;
		double new_x, Distance, new_y;
		double start_x, start_y;

		d_width = DisplayWidth( display, 0 );
		d_height = DisplayHeight( display, 0 );
		if ( ! XGetWindowAttributes( display, window, &watt ))
		{	fprintf( stderr, "XGetWindowAttributes failed\n" );
			exit ( 99 );
		}
		w_width = watt.width;
		w_height = watt.height;

		start_y = 0;
		start_x = d_width / 2;
			/* Start the following loop at the middle of the top of
			** the screen.
			*/
		start_x -= w_width / 2;
			/* Try to center the window. */

		d_width -= w_width + 10;
		d_height -= w_height + 10;
			/* allow a few pixels of slop, for borders and the like.
			*/

		Distance = (double)(((double)( d_width + d_height )) / 360.0 );
			/* There are 720 stops, and the total distance around
			** the screen is 2 * ( d_width + d_height );
			** so the Distance is ( 2 / 720 ) * ( d_width + d_height );
			*/

		for ( hour = minute = 0;; minute++ )
		{	/* Where should the window be at any given time? */
			if ( hour >= 11 && minute >= 60 ) break;
			if ( minute >= 60 )
			{	/* Next hour. */
				minute = 0;
				++hour;
			}

			new_y = start_y;
			new_x = start_x +
				((double)((double)(hour * 60 + minute )) * Distance );
			if ( new_x > d_width )
			{	/* Wrap around the top right corner. */
				new_y = new_x - (double)d_width;
				new_x = d_width;
				if ( new_y > d_height )
				{	/* Wrap around bottom right */
					new_x = (double)d_width -
						(double)( new_y - (double)d_height );
					new_y = d_height;
					if ( new_x < 0.0 )
					{	/* Wrap around bottom left */
						new_y = (double)d_height + new_x;
						new_x = 0.0;
						if ( new_y < 0.0 )
						{	/* Wrap top left */
							new_x = -new_y;
							new_y = 0;
							if ( new_x > start_x )
							{	/* oops. */
fprintf( stderr, "OOPS %d %d %Lf %Lf\n", hour, minute, new_x, new_y );
								exit ( 90 );
							}
						}
					}
				}
			}


/*
fprintf( stderr, "OK %d %d %Lf %Lf\n", hour, minute, new_x, new_y );
*/
			places[ hour ][ minute ].x = new_x;
			places[ hour ][ minute ].y = new_y;
		}

		for ( ;; )
		{	/* Move the window to its clockish position. */
			time_t ClocK;
			struct tm * times;

			time( &ClocK );
  			times = localtime( &ClocK );
  			hour = times->tm_hour % 12;
  			minute = times->tm_min;

			if ( ! XGetWindowAttributes( display, window, &watt ))
			{	/* In case the clock gets killed. */
				exit ( 98 );
			}

			XMoveWindow( display, window,
				places[ hour ][ minute ].x,
				places[ hour ][ minute ].y );
			XFlush( display );

/*
fprintf( stderr, "Moved %d ( 0x%x ) to %d %d, %d:%02d\n",
				window, window,
				places[ hour ][ minute ].x,
				places[ hour ][ minute ].y,
				hour, minute );
fflush( stderr );
*/

			sleep( 60 );
				/* sleep( 60 ); is imprecise, but good enough. */
		}
	}
}

char *help_text[] = {
"xMoveIt is a small tool to move another window around.",
"The syntax is:",
"",
"    xMoveIt [-display display][-name name] [-id id]",
"",
"the -name option chooses the target window by name,",
"the -id option chooses the target window by window-ID,",
"the -display option is standard X usage.",
"",
"If neither -id nor -name is specified,",
"xMoveIt prompts you to use the pointer to select a window.",
0};

void local_help()
{
    register char **ptr;

    for (ptr = help_text; *ptr; ptr++)
      fprintf(stderr,"%s\n",*ptr);

    exit(3);
}


/* Code for the next two routines was lifted from dsimple.c in
   xwininfo */

Window window_by_name(wdw,name)
Window wdw;
char *name;
{
    Window *offspring;		/* Any children */
    Window junk;		/* Just that */
    Window w = 0;		/* Found window */
    int count;			/* Number of kids */
    int loop;			/* Loop counter */
    char *wdw_name;		/* Returnewd name */

    if (XFetchName(display,wdw,&wdw_name) && !strcmp(wdw_name,name))
      return(wdw);

    if (!XQueryTree(display,wdw,&junk,&junk,&offspring,&count))
      return(0);

    for (loop = 0; loop < count; loop++)
      {
	  w = window_by_name(offspring[loop],name);
	  if (w)
	    break;
      }

    if (offspring)
      XFree(offspring);

    return(w);
}

Window point_to_window()
{
    int status;
    Cursor cursor;
    XEvent event;
    Window target_win = None;
    int buttons = 0;

    /* Make the target cursor */
    cursor = XCreateFontCursor(display, XC_crosshair);

    /* Grab the pointer using target cursor, letting it room all over */
    status = XGrabPointer(display, RootWindow(display, screen), False,
			  ButtonPressMask|ButtonReleaseMask, GrabModeSync,
			  GrabModeAsync, None, cursor, CurrentTime);
    if (status != GrabSuccess)
      {
	  fprintf(stderr,"Can't grab the mouse.");
	  exit(4);
      }

    /* Let the user select a window... */
    while ((target_win == None) || (buttons != 0)) /* allow one more event */
      {
	  XAllowEvents(display, SyncPointer, CurrentTime);
	  XWindowEvent(display, RootWindow(display, screen),
		       ButtonPressMask|ButtonReleaseMask, &event);
	  switch (event.type)
	    {
	      case ButtonPress:
		if (target_win == None)
		  {
		      /* window selected */
		      target_win = event.xbutton.subwindow;
		      if (target_win == None)
			target_win = RootWindow(display,screen);
		  }
		buttons++;
		break;

		/* there may have been some down before we started */
	      case ButtonRelease:
		if (buttons > 0)
		  buttons--;
		break;
	    }
      }

    XUngrabPointer(display, CurrentTime);      /* Done with pointer */

    return(target_win);
}
